#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Optimized kernel using shared memory and warp-level primitives for reductions

// Kernel applies: 1) F.linear(x, weight, bias)
//                2) F.max_pool1d on each kernel_size segment of the resulting vector
//                3) Sum of pooled max values
//                4) Scale the sum by scale_factor
// Each block processes one sample (batch element) and employs multiple threads

__global__ void optimized_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,      // batch size
    int inF,    // number of input features
    int outF,   // number of output features
    int kernel_size,
    float scale_factor
) {
    int b = blockIdx.x;
    if (b >= B) return;

    // Pointer to shared memory: used to store the linear layer outputs (size = outF)
    extern __shared__ float sdata[];  // sdata[0..outF-1]

    // Step 1: Compute F.linear: for each output feature j, compute dot product of x[b] and weight[j] plus bias
    // Distribute the computation across threads in the block
    for (int j = threadIdx.x; j < outF; j += blockDim.x) {
        float acc = bias[j];
        int x_offset = b * inF;
        int w_offset = j * inF;
        for (int i = 0; i < inF; i++) {
            acc += x[x_offset + i] * weight[w_offset + i];
        }
        sdata[j] = acc;
    }
    __syncthreads();

    // Step 2: Pooling - divide the output vector (of length outF) into segments of length 'kernel_size'
    // and compute the maximum within each segment
    int pooled_len = 0;
    if (outF >= kernel_size) {
        pooled_len = 1 + (outF - kernel_size) / kernel_size;
    }

    // Each thread processes multiple pooling segments in a strided loop
    float thread_pool_sum = 0.0f;
    for (int seg = threadIdx.x; seg < pooled_len; seg += blockDim.x) {
        int start = seg * kernel_size;
        float max_val = sdata[start];
        for (int j = 1; j < kernel_size; j++) {
            float tmp = sdata[start + j];
            if (tmp > max_val) {
                max_val = tmp;
            }
        }
        thread_pool_sum += max_val;
    }

    // Step 3: Intra-block reduction of thread_pool_sum using warp-level primitives
    // First, perform warp-level reduction using __shfl_down_sync
    unsigned int mask = 0xffffffff;
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        thread_pool_sum += __shfl_down_sync(mask, thread_pool_sum, offset);
    }

    // Each warp's lane 0 will store its reduced sum into shared memory
    __shared__ float warp_sums[32];  // sufficient for up to 1024 threads (32 warps)
    int lane = threadIdx.x & (warpSize - 1);
    int warpId = threadIdx.x / warpSize;
    if (lane == 0) {
        warp_sums[warpId] = thread_pool_sum;
    }
    __syncthreads();

    // Final reduction: use the first warp to reduce the sums from each warp
    // Compute number of warps participating
    int numWarps = blockDim.x / warpSize;
    float final_sum = 0.0f;
    if (threadIdx.x < 32) {
        // Load value if this thread is within the number of warps, else set to 0
        float val = (threadIdx.x < numWarps) ? warp_sums[threadIdx.x] : 0.0f;
        for (int offset = 16; offset > 0; offset /= 2) {
            val += __shfl_down_sync(mask, val, offset);
        }
        if (threadIdx.x == 0) {
            final_sum = val;
            // Step 4: Scale the result and write output
            output[b] = final_sum * scale_factor;
        }
    }
}

// Forward function callable from Python
at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    const auto B = x.size(0);
    const auto inF = x.size(1);
    const auto outF = weight.size(0);

    // Prepare output tensor
    auto out = torch::empty({B}, x.options());

    // Launch the kernel: one block per sample, 256 threads per block
    // Allocate shared memory equal to outF * sizeof(float) for storing intermediate F.linear outputs
    int threads = 256;
    size_t sharedMemSize = outF * sizeof(float);
    optimized_kernel<<<B, threads, sharedMemSize>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

// PyBind11 module registration
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "CUDA forward for optimized_kernel");
}
